import os
import re
import time
from typing import Optional, List
import glob

from unidecode import unidecode

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver

from argparse import ArgumentParser, Namespace

from ontol import (
    Parser,
    JSONSerializer,
    PlantUML,
    Retranslator,
    Ontology,
    Figure,
    AI,
    constants,
)

__VERSION__ = "2.14.1"


class CLI:
    def __init__(self) -> None:
        self.args_parser: ArgumentParser = ArgumentParser(
            description='Ontol DSL Parser - A tool for parsing and visualizing ontology files written in the Ontol DSL.'
        )
        self.args_parser.add_argument(
            'file',
            type=str,
            help='Path to the .ontol file to be parsed. Supports wildcards and directories.',
        )
        self.args_parser.add_argument(
            '-w',
            '--watch',
            action='store_true',
            default=False,
            help='Watch the specified file(s) for changes and re-parse automatically',
        )
        self.args_parser.add_argument(
            '-d',
            '--debug',
            action='store_true',
            default=False,
            help='Get retranslation version of the file .ontol',
        )
        self.args_parser.add_argument(
            '-q',
            '--quiet',
            action='store_true',
            default=False,
            help='Ignore all the warnings',
        )
        self.args_parser.add_argument(
            '--output-dir',
            dest='output_dir',
            type=str,
            help='Output directory to write files in',
        )
        self.args_parser.add_argument(
            '--gen-hierarchy',
            dest='gen_hierarchy',
            action='store_true',
            default=False,
            help='Generate additional hierarchy relationship using Ollama model',
        )
        self.args_parser.add_argument(
            '-m',
            '--model',
            type=str,
            default='llama3.1',
            help='Ollama model to use',
        )
        self.args_parser.add_argument(
            '-t',
            '--temperature',
            type=float,
            default=0.0,
            help='Model temperature to use',
        )
        self.args_parser.add_argument(
            '--split-funcs-rels',
            dest='split_funcs_rels',
            action='store_true',
            default=False,
            help='Split functions and relations',
        )
        self.args_parser.add_argument(
            '-v',
            '--version',
            action='version',
            version=f'%(prog)s {__VERSION__}',
            help='Show the version of the program and exit',
        )
        self.args_parser.add_argument(
            '--max-edges',
            dest='max_edges',
            type=int,
            help='Set max edges in scheme',
        )

        self.parser: Parser = Parser()
        self.serializer: JSONSerializer = JSONSerializer()
        self.plantuml: PlantUML = PlantUML()
        self.retranslator: Retranslator = Retranslator()
        self.ai: AI = AI()

    def run(self) -> None:
        args: Namespace = self.args_parser.parse_args()

        file_paths = self.get_file_paths(args.file)
        # Ensure we attempt to parse even non-existent files
        if not file_paths:
            file_paths = [args.file]

        if args.watch:
            for file_path in file_paths:
                self.watch_file(file_path, args)
        else:
            for file_path in file_paths:
                self.parse_file(file_path, args)

    def get_file_paths(self, path: str) -> List[str]:
        # Check if the path is a directory
        if os.path.isdir(path):
            # Recursively find all .ontol files
            ontol_files = []
            for root, _, files in os.walk(path):
                for file in files:
                    if file.endswith('.ontol'):
                        ontol_files.append(os.path.join(root, file))
            return ontol_files
        # If not a directory, assume it's a pattern or a file
        else:
            return glob.glob(path, recursive=True)

    def parse_file(self, file_path: str, args: Optional[Namespace] = None) -> None:
        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                content: str = file.read()
                ontology, warnings = self.parser.parse(content, file_path)

                if (
                    args
                    and args.max_edges
                    and (count := ontology.count_edges()) > args.max_edges
                ):
                    warnings.append(
                        f'{constants.warning_prefix} too much edges. Expected: '
                        f'{args.max_edges}, got: {count}'
                    )

                # Print warnings
                if warnings and (not args or not args.quiet):
                    print('\n\n'.join(warnings))

                if args and args.gen_hierarchy:
                    print('Generating hierarchy...')
                    hierarchy, comments = self.ai.generate_hierarchy(
                        ontology, args.model, args.temperature
                    )
                    ontology.hierarchy.extend(hierarchy)
                    print('\nGenerated relationships:')
                    for relationship, comment in zip(hierarchy, comments):
                        print(
                            f'{relationship.parent.name} {relationship.relationship.value} {relationship.children[0].name}: {comment}'
                        )

                base_dir = os.path.dirname(file_path)
                base_name = os.path.splitext(os.path.basename(file_path))[0]
                output_dir = args.output_dir if args and args.output_dir else base_dir

                if output_dir:
                    os.makedirs(output_dir, exist_ok=True)

                ontologies: list[Ontology] = [ontology]
                base_names: list[str] = [base_name]

                for figure in ontology.figures:
                    ontologies.append(Ontology.from_figure(ontology, figure))
                    file_postfix = self.get_figure_file_postfix(figure)
                    base_names.append(f'{base_name}_{file_postfix}')

                if args and args.split_funcs_rels:
                    new_ontologies: list[Ontology] = []
                    new_base_names: list[str] = []

                    for ontology, base_name in zip(ontologies, base_names):
                        new_ontologies.extend(
                            [ontology.without_functions, ontology.only_functions]
                        )
                        new_base_names.extend(
                            [f'{base_name}_rels', f'{base_name}_funcs']
                        )

                    ontologies.extend(new_ontologies)
                    base_names.extend(new_base_names)

                for ontology, base_name in zip(ontologies, base_names):
                    # JSON
                    json_content: str = self.serializer.serialize(ontology)
                    json_file_path: str = os.path.join(output_dir, f'{base_name}.json')
                    with open(json_file_path, 'w', encoding='utf-8') as json_file:
                        json_file.write(json_content)

                    # PlantUML
                    plantuml_content: str = self.plantuml.generate(ontology)
                    puml_file_path: str = os.path.join(output_dir, f'{base_name}.puml')
                    with open(puml_file_path, 'w', encoding='utf-8') as puml_file:
                        puml_file.write(plantuml_content)
                    self.plantuml.processes_puml_to_png(puml_file_path)

                    # Retranslator
                    if args and not args.debug:
                        continue
                    retranslator_content: str = self.retranslator.translate(ontology)
                    retranslator_file_path: str = os.path.join(
                        output_dir, f'{base_name}_retr.ontol'
                    )
                    with open(
                        retranslator_file_path, 'w', encoding='utf-8'
                    ) as retr_file:
                        retr_file.write(retranslator_content)
        except FileNotFoundError:
            print(f"{constants.error_prefix} the file '{file_path}' does not exist.")
        except Exception as e:
            print(f'{constants.error_prefix} error processing file {file_path}: {e}')

    def watch_file(self, file_path: str, args: Optional[Namespace] = None):
        self.parse_file(file_path, args)

        class FileChangeHandler(FileSystemEventHandler):
            def __init__(self, parse_callback):
                super().__init__()
                self.parse_callback = parse_callback

            def on_modified(self, event):
                if event.src_path.endswith('.ontol'):
                    print(f'File {event.src_path} modified, re-parsing...')
                    self.parse_callback(event.src_path, args)

        event_handler: FileChangeHandler = FileChangeHandler(self.parse_file)

        observer: BaseObserver = Observer()
        observer.schedule(event_handler, path=file_path, recursive=True)
        observer.start()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

    def get_figure_file_postfix(self, figure: Figure) -> str:
        file_postfix: str = figure.name.strip().lower()
        file_postfix = unidecode(file_postfix)
        file_postfix = re.sub(r'[^a-zA-Z0-9\s]', ' ', file_postfix)
        file_postfix = re.sub(r'\s+', ' ', file_postfix)
        file_postfix = re.sub(r'\s', '_', file_postfix)
        return file_postfix


def main():
    cli: CLI = CLI()
    cli.run()


if __name__ == '__main__':
    main()



from ontol import (
    Function,
    Meta,
    Ontology,
    Relationship,
    Term,
    FunctionArgument,
    RelationshipType,
    TermAttributes,
)


def test_term_creation() -> None:
    term: Term = Term(name='test_term', label='TestTerm', description='A test term')
    assert term.name == 'test_term'
    assert term.label == 'TestTerm'
    assert term.description == 'A test term'
    assert (
        repr(term)
        == 'Term(name=test_term, label=TestTerm, description=A test term, attributes=TermAttributes(color=None, note=None))'
    )


def test_term_with_attributes_creation() -> None:
    attributes = TermAttributes(color='#red')
    term_with_attributes: Term = Term(
        name='test_term',
        label='TestTerm',
        description='A test term',
        attributes=attributes,
    )
    assert term_with_attributes.attributes == attributes
    assert (
        repr(term_with_attributes)
        == f'Term(name=test_term, label=TestTerm, description=A test term, attributes={attributes})'
    )


def test_term_with_empty_label_creation() -> None:
    term: Term = Term(name='test_term', label='', description='A test term')
    assert term.label == ''
    assert (
        repr(term)
        == 'Term(name=test_term, label=, description=A test term, attributes=TermAttributes(color=None, note=None'
        '))'
    )


def test_term_with_empty_description_creation() -> None:
    term: Term = Term(name='test_term', label='TestTerm', description='')
    assert term.description == ''
    assert (
        repr(term)
        == 'Term(name=test_term, label=TestTerm, description=, attributes=TermAttributes(color=None, note=None'
        '))'
    )


def test_function_creation():
    func: Function = Function(
        name='sum',
        label='Sum',
        input_types=[
            FunctionArgument(Term('int'), 'First number'),
            FunctionArgument(Term('int'), 'Second number'),
        ],
        output_type=FunctionArgument(Term('int'), 'Result'),
    )
    assert func.name == 'sum'
    assert func.label == 'Sum'
    assert func.input_types == [
        FunctionArgument(Term('int'), 'First number'),
        FunctionArgument(Term('int'), 'Second number'),
    ]
    assert func.output_type == FunctionArgument(Term('int'), 'Result')
    assert (
        repr(func)
        == "Function(name=sum, label=Sum, input_types=[('int', 'First number'), ('int', 'Second number')], output_type=('int', 'Result'), attributes=FunctionAttributes(color=None, colorArrow=None, type=None, inputTitle=None, outputTitle=None))"
    )


def test_function_wth_empty_label_creation():
    func: Function = Function(
        name='sum',
        label='',
        input_types=[
            FunctionArgument(Term('int'), 'First number'),
            FunctionArgument(Term('int'), 'Second number'),
        ],
        output_type=FunctionArgument(Term('int'), 'Result'),
    )
    assert func.label == ''
    assert (
        repr(func)
        == "Function(name=sum, label=, input_types=[('int', 'First number'), ('int', 'Second number')], output_type=('int', 'Result'), attributes=FunctionAttributes(color=None, colorArrow=None, type=None, inputTitle=None, outputTitle=None))"
    )


def test_function_wth_empty_input_types_creation():
    func: Function = Function(
        name='sum',
        label='Sum',
        input_types=[],
        output_type=FunctionArgument(Term('int'), 'Result'),
    )
    assert func.input_types == []
    assert (
        repr(func) == 'Function(name=sum, label=Sum, input_types=[],'
        " output_type=('int', 'Result'), attributes=FunctionAttributes(color=None, colorArrow=None, type=None, inputTitle=None, outputTitle=None))"
    )


def test_relationship_creation() -> None:
    rel: Relationship = Relationship(
        parent=Term('set'),
        relationship=RelationshipType.DEPENDENCE,
        children=[Term('element')],
    )
    assert rel.parent.name == 'set'
    assert rel.relationship.value == 'dependence'
    assert rel.children == [Term('element')]
    assert (
        repr(rel)
        == 'Relationship(parent=set, relationship=dependence, children=[element], attributes=RelationshipAttributes(color=None, direction=None, title=None, rightChar=None, leftChar=None))'
    )


def test_meta_creation():
    meta: Meta = Meta(
        version='1.0',
        title='TestOntology',
        author='Author',
        description='A test ontology',
        type='Base',
        date='2024-01-01',
    )
    assert meta.version == '1.0'
    assert meta.title == 'TestOntology'
    assert meta.author == 'Author'
    assert meta.description == 'A test ontology'
    assert meta.type == 'Base'
    assert meta.date == '2024-01-01'
    assert (
        repr(meta) == 'Meta(version=1.0, title=TestOntology, author=Author,'
        ' description=A test ontology, type=Base, date=2024-01-01)'
    )


def test_meta_with_empty_description_creation():
    meta_no_optional: Meta = Meta(
        version='1.0',
        title='TestOntology',
        author='Author',
        description=None,
        type='Base',
        date='2024-01-01',
    )
    assert meta_no_optional.description is None
    assert (
        repr(meta_no_optional)
        == 'Meta(version=1.0, title=TestOntology, author=Author, description=None, type=Base, date=2024-01-01)'
    )


def test_ontology_operations():
    ontology: Ontology = Ontology()
    char: Term = Term(name='str', label='String', description='')
    string: Term = Term(name='string', label='String', description='')
    concatenate: Function = Function(
        name='concatenate',
        label='Concatenate',
        input_types=[
            {'name': Term(name='str', label='String', description=''), 'label': ''}
        ],
        output_type=(
            {
                'name': Term(name='str', label='String', description=''),
                'label': 'Result',
            }
        ),
    )
    rel: Relationship = Relationship(
        parent=Term(name='string', label='String', description=''),
        relationship=RelationshipType.COMPOSITION,
        children=[Term(name='str', label='String', description='')],
    )
    meta: Meta = Meta(
        version='1.0',
        title='StringOntology',
        author='Author',
        description='A string ontology',
        date='2024-01-01',
    )

    ontology.add_type(char)
    ontology.add_type(string)
    ontology.add_function(concatenate)
    ontology.add_relationship(rel)
    ontology.set_meta(meta)

    assert ontology.types == [char, string]
    assert ontology.functions == [concatenate]
    assert ontology.hierarchy == [rel]
    assert ontology.meta == meta

    assert repr([char, string]) in repr(ontology)
    assert repr([concatenate]) in repr(ontology)
    assert repr([rel]) in repr(ontology)
    assert repr(meta) in repr(ontology)



import json
from typing import Any

from dataclasses import asdict

from ontol import Ontology, Term, Function, Meta, Relationship, Figure


class JSONSerializer:
    @staticmethod
    def serialize(ontology: Ontology) -> str:
        data: dict[str, object] = {
            'meta': JSONSerializer._serialize_meta(ontology.meta),
            'terms': [JSONSerializer._serialize_term(t) for t in ontology.types],
            'functions': [
                JSONSerializer._serialize_function(f) for f in ontology.functions
            ],
            'hierarchy': [
                JSONSerializer._serialize_relationship(r) for r in ontology.hierarchy
            ],
            'figures': [JSONSerializer._serialize_figure(f) for f in ontology.figures],
        }
        return json.dumps(data, ensure_ascii=False, indent=4)

    @staticmethod
    def _serialize_term(type_def: Term) -> dict[str, Any]:
        return {
            'name': type_def.name,
            'label': type_def.label,
            'description': type_def.description,
            'attributes': {
                key: value
                for key, value in asdict(type_def.attributes).items()
                if value is not None
            },
        }

    @staticmethod
    def _serialize_function(
        func_def: Function,
    ) -> dict[str, Any]:
        attributes = {
            key: value
            for key, value in asdict(func_def.attributes).items()
            if value is not None
        }

        if 'type' in attributes:
            attributes['type'] = attributes['type'].value

        return {
            'name': func_def.name,
            'label': func_def.label,
            'input_types': [
                {'name': t.term.name, 'label': t.label} for t in func_def.input_types
            ],
            'output_type': {
                'name': func_def.output_type.term.name,
                'label': func_def.output_type.label,
            },
            'attributes': attributes,
        }

    @staticmethod
    def _serialize_meta(meta: Meta) -> dict[str, str | None]:
        return asdict(meta)

    @staticmethod
    def _serialize_relationship(
        rel_def: Relationship,
    ) -> dict[str, Any]:
        attributes = {
            key: value
            for key, value in asdict(rel_def.attributes).items()
            if value is not None
        }

        if 'direction' in attributes:
            attributes['direction'] = attributes['direction'].value

        return {
            'name': rel_def.name,
            'parent': rel_def.parent.name,
            'relationship': rel_def.relationship.value,
            'children': [child.name for child in rel_def.children],
            'attributes': attributes,
        }

    @staticmethod
    def _serialize_figure(
        figure: Figure,
    ) -> dict[str, Any]:
        return {
            'name': figure.name,
            'terms': [term.name for term in figure.types],
            'functions': [func.name for func in figure.functions],
            'hierarchy': [rel.name for rel in figure.hierarchy],
        }



from ontol import (
    Function,
    Meta,
    Ontology,
    Relationship,
    Term,
    FunctionArgument,
    PlantUML,
    RelationshipType,
    TermAttributes,
    FunctionAttributes,
)

import pytest


@pytest.fixture
def mock_ontology():
    ontology = Ontology()
    ontology.set_meta(
        Meta(
            version='1.0',
            title='TestOntology',
            author='Author',
            description='Test ontology',
            type='Базовый',
            date='2024-01-01',
        )
    )
    # TODO: add term without color attribute
    ontology.add_type(
        Term(
            name='MyTypeParent',
            label='test label Term1',
            description='A test type1',
            attributes=TermAttributes(color='#E6B8B7'),
        )
    )
    ontology.add_type(
        Term(name='MyTypeChild', label='test label Term2', description='A test type2')
    )
    ontology.add_function(
        Function(
            name='MyFunction1',
            label='test label Func1',
            input_types=[
                FunctionArgument(Term('MyTypeChild'), 'test1 in1'),
                FunctionArgument(Term('MyTypeChild'), 'test1 in2'),
            ],
            output_type=FunctionArgument(Term('MyTypeParent'), 'test_type1'),
            attributes=FunctionAttributes(colorArrow='#E6B8B7'),
        )
    )
    ontology.add_function(
        Function(
            name='MyFunction2',
            label='test label Func2',
            input_types=[FunctionArgument(Term('MyTypeChild'), 'test2 in1')],
            output_type=FunctionArgument(Term('MyTypeParent'), 'test_type2'),
        )
    )
    ontology.add_relationship(
        Relationship(
            parent=Term('MyTypeParent'),
            relationship=RelationshipType.COMPOSITION,
            children=[Term('MyTypeChild')],
        )
    )
    return ontology


@pytest.fixture
def generator():
    return PlantUML()


def test_generate_full_uml(generator: PlantUML, mock_ontology):
    uml_output = generator.generate(mock_ontology)
    assert '@startuml' in uml_output
    assert '@enduml' in uml_output

    assert 'MyTypeParent' in uml_output
    assert 'test label Term1' in uml_output
    assert '(A test type1)' in uml_output
    assert '#E6B8B7' in uml_output
    assert 'MyTypeChild' in uml_output
    assert 'test label Term2' in uml_output
    assert '(A test type2)' in uml_output
    assert '#white' in uml_output

    assert (
        '"test label Func1\\n(MyTypeChild: test1 in1, MyTypeChild: test1 in2 -> MyTypeParent: test_type1)"'
        ' as MyFunction1 #white' in uml_output
    )
    assert (
        '"test label Func2\\n(MyTypeChild: test2 in1 -> MyTypeParent: test_type2)" as MyFunction2 #white'
        in uml_output
    )
    print(uml_output)
    assert 'MyTypeChild "2" --[#E6B8B7]->  MyFunction1 ' in uml_output
    assert 'MyTypeChild  --[#black]->  MyFunction2 ' in uml_output
    assert 'MyFunction1  --[#E6B8B7]->  MyTypeParent ' in uml_output
    assert 'MyFunction2  --[#black]->  MyTypeParent ' in uml_output

    assert 'MyTypeParent  --[#black]-*  MyTypeChild ' in uml_output


def test_generate_type(generator: PlantUML):
    term = Term(name='MyType', description='A test type', label='test label')
    result = generator._generate_type(term)
    assert result == 'class MyType {\n  A test type\n}'


def test_generate_function(generator: PlantUML):
    func = Function(
        name='MyFunction',
        label='test label',
        input_types=[FunctionArgument(Term('int'), 'test')],
        output_type=FunctionArgument(Term('bool'), 'test'),
        attributes={'color': '#E6B8B7'},
    )
    result = generator._generate_function(func)
    assert 'MyFunction' in result
    assert 'class MyFunction <<Function>> {' in result
    assert '+MyFunction(int) : (bool)' in result
    assert 'test label' in result
    assert '}' in result


def test_generate_relationship(generator: PlantUML):
    relation = Relationship(
        parent=Term('MyTypeParent'),
        relationship=RelationshipType.from_str('aggregation'),
        children=[Term('MyTypeChild')],
    )
    result = generator._generate_relationship(relation)
    assert 'MyTypeParent' in result
    assert 'aggregation' in result
    assert 'MyTypeChild' in result
    assert 'as' in result



from ontol import (
    Function,
    Meta,
    Ontology,
    Relationship,
    Term,
    FunctionArgument,
    RelationshipType,
    TermAttributes,
    FunctionAttributes,
    Retranslator,
)

import pytest


@pytest.fixture
def mock_ontology():
    ontology = Ontology()
    ontology.set_meta(
        Meta(
            version='1.0',
            title='TestOntology',
            author='Author',
            description='Test ontology',
            type='Базовый',
            date='2024-01-01',
        )
    )
    ontology.add_type(
        Term(
            name='MyTypeParent',
            label='test label Term1',
            description='A test type1',
            attributes=TermAttributes(color='#E6B8B7'),
        )
    )
    ontology.add_type(
        Term(name='MyTypeChild', label='test label Term2', description='A test type2')
    )
    ontology.add_function(
        Function(
            name='MyFunction1',
            label='test label Func1',
            input_types=[
                FunctionArgument(Term('MyTypeChild'), 'test1 in1'),
                FunctionArgument(Term('MyTypeChild'), 'test1 in2'),
            ],
            output_type=FunctionArgument(Term('MyTypeParent'), 'test_type1'),
            attributes=FunctionAttributes(colorArrow='#E6B8B7'),
        )
    )
    ontology.add_relationship(
        Relationship(
            parent=Term('MyTypeParent'),
            relationship=RelationshipType.COMPOSITION,
            children=[Term('MyTypeChild')],
        )
    )
    return ontology


@pytest.fixture
def retranslator():
    return Retranslator()


def test_translate_full_ontology(retranslator: Retranslator, mock_ontology):
    retr_output = retranslator.translate(mock_ontology)
    assert "version: '1.0'" in retr_output
    assert "title: 'TestOntology'" in retr_output
    assert "author: 'Author'" in retr_output
    assert "description: 'Test ontology'" in retr_output
    assert "type: 'Базовый'" in retr_output
    assert "date: '2024-01-01'" in retr_output

    assert 'types:' in retr_output
    assert "MyTypeParent: 'test label Term1', 'A test type1', { color: '#E6B8B7' }" in retr_output
    assert "MyTypeChild: 'test label Term2', 'A test type2'" in retr_output

    assert 'functions:' in retr_output
    assert ("MyFunction1: 'test label Func1' "
            "(MyTypeChild: 'test1 in1', MyTypeChild: 'test1 in2')"
            " -> MyTypeParent: 'test_type1', { colorArrow: '#E6B8B7' }") in retr_output

    assert 'hierarchy:' in retr_output
    assert 'MyTypeParent composition MyTypeChild'



import json
from dataclasses import fields
from typing import Union

from ontol import (
    Function,
    Ontology,
    Relationship,
    Term,
    RelationshipAttributes,
    TermAttributes,
    FunctionAttributes,
)


class Retranslator:
    def translate(self, ontology: Ontology) -> str:
        ontol_lines: list[str] = []

        for field in fields(ontology.meta):
            value = getattr(ontology.meta, field.name)
            if value is not None:
                ontol_lines.append(f'{field.name}: {value!r}')
        ontol_lines.append('')

        if ontology.types:
            ontol_lines.append('types:')
            for term in ontology.types:
                ontol_lines.append(self.__translate_term(term))
            ontol_lines.append('')

        if ontology.functions:
            ontol_lines.append('functions:')
            for function in ontology.functions:
                ontol_lines.append(self.__translate_function(function))
            ontol_lines.append('')

        if ontology.hierarchy:
            ontol_lines.append('hierarchy:')
            for relationship in ontology.hierarchy:
                ontol_lines.append(self.__translate_hierarchy(relationship))
            ontol_lines.append('')

        if ontology.figures:
            for figure in ontology.figures:
                ontol_lines.append(f'figure {figure.name}:')
                for definition in figure.types + figure.functions + figure.hierarchy:
                    if definition.name is not None:
                        ontol_lines.append(definition.name)
                ontol_lines.append('')

        return '\n'.join(ontol_lines)

    def __translate_term(self, term: Term) -> str:
        label = "'" + term.label + "'"
        term_line = f'{term.name}: {label}, {term.description!r}'

        return term_line + self.__translate_attributes(term.attributes)

    def __translate_function(self, function: Function) -> str:
        label = "'" + function.label + "'"
        function_line = f'{function.name}: {label} ('
        args = ''
        for i, arg in enumerate(function.input_types):
            args += f'{arg.term.name}: {arg.label!r}'
            if i < len(function.input_types) - 1:
                args += ', '
        function_line += (
            f'{args}) -> '
            f'{function.output_type.term.name}: {function.output_type.label!r}'
        )

        return function_line + self.__translate_attributes(function.attributes)

    def __translate_hierarchy(self, relationship: Relationship) -> str:
        relationship_line = (
            f'{relationship.parent.name} '
            f'{relationship.relationship.value} '
            f'{relationship.children[0].name}'
        )

        return relationship_line + self.__translate_attributes(relationship.attributes)

    def __translate_attributes(
        self,
        attributes: Union[RelationshipAttributes, FunctionAttributes, TermAttributes],
    ) -> str:
        res, non_none_fields = '', []
        for field in fields(attributes):
            if (attr := getattr(attributes, field.name)) is not None:
                if isinstance(attr, str):
                    attribute = attr
                else:
                    attribute = attr.value
                non_none_fields.append(f'{field.name}: {attribute!r} ')

        if non_none_fields:
            res += ', { ' + ', '.join(non_none_fields) + '}'

        return res



import collections
import os
import zlib
import requests
from typing import Optional

from ontol import (
    Function,
    Ontology,
    Relationship,
    Term,
    RelationshipDirection,
    RelationshipType,
    RelationshipAttributes,
    TermAttributes,
)


# TODO: make look like in technical task
class PlantUML:
    SERVER_URL: str = 'http://www.plantuml.com/plantuml/png/'

    def __init__(self, url=SERVER_URL):
        self.url = url

    def generate(self, ontology: Ontology) -> str:
        return self._generate_base(ontology)

    def _generate_base(self, ontology: Ontology) -> str:
        uml_lines: list[str] = [
            '@startuml',
            'skinparam backgroundColor #F0F8FF',
            'skinparam defaultTextAlignment center',
            'skinparam shadowing false',
            'skinparam dpi 150',
            'skinparam linetype ortho',
            'skinparam ranksep 40',
            'skinparam nodesep 30',
            f'package "'
            f'{ontology.meta.title if ontology.meta.title is not None else "Онтология"}'
            f'" {{',
        ]

        for term in ontology.types:
            uml_lines.append(self._generate_rectangle(term))
            uml_lines.append(self._generate_note(term))

        for function in ontology.functions:
            uml_lines.append(
                self._generate_rectangle(self.__prepare_function_term(function))
            )

        for function in ontology.functions:
            for relations in self.__prepare_function_hierarchy(function, ontology):
                uml_lines.append(self._generate_base_hierarchy(relations))

        for relationship in ontology.hierarchy:
            uml_lines.append(self._generate_base_hierarchy(relationship))

        uml_lines.append('}')
        uml_lines.append('@enduml')
        return '\n'.join(uml_lines)

    @staticmethod
    def _generate_rectangle(term: Term) -> str:
        return (
            f'rectangle "{term.label}'
            + (f'\\n({term.description})' if term.description else '')
            + f'" as {term.name} {term.attributes.color or "#white"}'
        )

    @staticmethod
    def _generate_note(term: Term) -> str:
        res = ''
        if term.attributes.note:
            note_text = term.attributes.note.replace('\\n', '\n')
            res = f'note right of {term.name}\n{note_text}\n    end note'
        return res

    @staticmethod
    def _generate_base_hierarchy(relationship: Relationship) -> str:
        relationships = {
            RelationshipType.DEPENDENCE.value: {
                RelationshipDirection.FORWARD.value: '...>',
                RelationshipDirection.BACKWARD.value: '<...',
                RelationshipDirection.BIDIRECTIONAL.value: '<...>',
            },
            RelationshipType.ASSOCIATION.value: {
                RelationshipDirection.FORWARD.value: '---',
                RelationshipDirection.BACKWARD.value: '---',
                RelationshipDirection.BIDIRECTIONAL.value: '---',
            },
            RelationshipType.DIRECT_ASSOCIATION.value: {
                RelationshipDirection.FORWARD.value: '--->',
                RelationshipDirection.BACKWARD.value: '<---',
                RelationshipDirection.BIDIRECTIONAL.value: '<--->',
            },
            RelationshipType.INHERITANCE.value: {
                RelationshipDirection.FORWARD.value: '---|>',
                RelationshipDirection.BACKWARD.value: '<|---',
                RelationshipDirection.BIDIRECTIONAL.value: '<|---|>',
            },
            RelationshipType.IMPLEMENTATION.value: {
                RelationshipDirection.FORWARD.value: '...|>',
                RelationshipDirection.BACKWARD.value: '<|...',
                RelationshipDirection.BIDIRECTIONAL.value: '<|...|>',
            },
            RelationshipType.AGGREGATION.value: {
                RelationshipDirection.FORWARD.value: '---o',
                RelationshipDirection.BACKWARD.value: 'o---',
                RelationshipDirection.BIDIRECTIONAL.value: 'o---o',
            },
            RelationshipType.COMPOSITION.value: {
                RelationshipDirection.FORWARD.value: '---*',
                RelationshipDirection.BACKWARD.value: '*---',
                RelationshipDirection.BIDIRECTIONAL.value: '*---*',
            },
        }
        leftchar: str = (
            ('"' + relationship.attributes.leftChar + '"')
            if relationship.attributes.leftChar
            else ''
        )
        rightchar: str = (
            ('"' + relationship.attributes.rightChar + '"')
            if relationship.attributes.rightChar
            else ''
        )
        title: str = (
            (': "' + relationship.attributes.title + '"')
            if relationship.attributes.title
            else ''
        )
        color: str = '[' + (relationship.attributes.color or '#black') + ']'
        slice = (
            3
            if relationship.attributes.direction == RelationshipDirection.BIDIRECTIONAL
            else 2
        )

        relation: str = (
            relationships[relationship.relationship.value][
                relationship.attributes.direction.value
                if relationship.attributes.direction
                else RelationshipDirection.FORWARD.value
            ][:slice]
            + color
            + relationships[relationship.relationship.value][
                relationship.attributes.direction.value
                if relationship.attributes.direction
                else RelationshipDirection.FORWARD.value
            ][slice:]
        )
        res: str = ''
        res += (
            f'{relationship.parent.name} {leftchar} '
            f'{relation} '
            f'{rightchar} '
            f'{relationship.children[0].name} {title}'
        ) + '\n'
        return res

    @staticmethod
    def __prepare_function_term(function: Function):
        input_str: list[str] = [
            f'{el.term.name}: {el.label}' if el.label else str(el.term.name)
            for el in function.input_types
        ]
        output_str: str = (
            f'{function.output_type.term.name}: {function.output_type.label}'
            if function.output_type.label
            else str(function.output_type.term.name)
        )
        description: str = f'{", ".join(input_str)} -> {output_str}'
        return Term(
            function.name,
            function.label,
            description,
            TermAttributes(color=function.attributes.color or '#white'),
        )

    @staticmethod
    def __prepare_function_hierarchy(function: Function, ontology: Ontology):
        relations = []
        input_types: collections.defaultdict[str, int] = collections.defaultdict(int)
        for input_type in function.input_types:
            input_types[input_type.term.name] += 1
        for k, v in input_types.items():
            term: Optional[Term] = ontology.find_term_by_name(k)
            if term is None:
                continue
            relations.append(
                Relationship(
                    parent=term,
                    relationship=RelationshipType.from_str(
                        function.attributes.type.value
                    )
                    if function.attributes.type
                    else RelationshipType.DIRECT_ASSOCIATION,
                    children=[Term(function.name)],
                    attributes=RelationshipAttributes(
                        color=function.attributes.colorArrow or '#black',
                        title=function.attributes.inputTitle or '',
                        leftChar=f'{v if v != 1 else ""}',
                        direction=RelationshipDirection.FORWARD,
                    ),
                )
            )
        relations.append(
            Relationship(
                parent=Term(function.name),
                relationship=RelationshipType.from_str(function.attributes.type.value)
                if function.attributes.type
                else RelationshipType.DIRECT_ASSOCIATION,
                children=[ontology.find_term_by_name(function.output_type.term.name)],
                attributes=RelationshipAttributes(
                    color=function.attributes.colorArrow,
                    title=function.attributes.outputTitle or '',
                    direction=RelationshipDirection.FORWARD,
                ),
            )
        )
        return relations

    def _generate_type(self, term: Term) -> str:
        return f'class {term.name} {{\n  {term.description}\n}}'

    # TODO: add comments to input and output types and add color for block
    def _generate_function(self, function: Function) -> str:
        inputs: str = ', '.join(map(lambda t: t.term.name, function.input_types))
        outputs: str = function.output_type.term.name
        return (
            f'class {function.name} <<Function>> {{\n'
            f'  +{function.name}({inputs}) : ({outputs})\n'
            f'  {function.label}\n}}'
        )

    # TODO: check type of relationship
    def _generate_relationship(self, relationship: Relationship) -> str:
        return (
            f'note "{relationship.parent.name} {relationship.relationship.value} {list(map(lambda t: t.name, relationship.children))}" '
            f'as N{hash(relationship.parent.name) % 10000}'
        )

    def processes_puml_to_png(self, puml_file):
        outfile = os.path.splitext(puml_file)[0] + '.png'

        with open(puml_file, 'r', encoding='utf-8') as file:
            plantuml_text = file.read()

        data = zlib.compress(plantuml_text.encode('utf-8'))[2:-4]
        encoded_text = ''
        for i in range(0, len(data), 3):
            if i + 2 == len(data):
                encoded_text += self.__encode3bytes(data[i], data[i + 1], 0)
            elif i + 1 == len(data):
                encoded_text += self.__encode3bytes(data[i], 0, 0)
            else:
                encoded_text += self.__encode3bytes(data[i], data[i + 1], data[i + 2])

        url = f'{self.url}{encoded_text}'
        response = requests.get(url)
        response.raise_for_status()

        if response.status_code == 200:
            with open(outfile, 'wb') as out:
                out.write(response.content)

    @staticmethod
    def __encode3bytes(b1, b2, b3):
        chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_'
        c1 = b1 >> 2
        c2 = ((b1 & 0x3) << 4) | (b2 >> 4)
        c3 = ((b2 & 0xF) << 2) | (b3 >> 6)
        c4 = b3 & 0x3F
        return chars[c1] + chars[c2] + chars[c3] + chars[c4]



import json

from ontol import (
    Function,
    Meta,
    Ontology,
    Relationship,
    Term,
    JSONSerializer,
    FunctionArgument,
    RelationshipType,
)

import pytest


@pytest.fixture
def sample_ontology():
    ontology = Ontology()
    ontology.add_type(
        Term(name='MyType', description='A sample type', label='A some label')
    )
    input_type: FunctionArgument = FunctionArgument(Term('int'))
    ontology.add_function(
        Function(
            name='MyFunction',
            label='Function',
            input_types=[input_type],
            output_type=FunctionArgument(Term('bool')),
        )
    )
    ontology.add_relationship(
        Relationship(
            relationship=RelationshipType.DEPENDENCE,
            parent=Term('int'),
            children=[Term('bool')],
        )
    )
    ontology.set_meta(
        Meta(
            version='1.0',
            title='SampleOntology',
            author='Author',
            description='Sample description',
            type='Базовый',
            date='2024-01-01',
        )
    )
    return ontology


@pytest.fixture
def serializer():
    return JSONSerializer()


def test_serialize_ontology(serializer, sample_ontology):
    json_output = serializer.serialize(sample_ontology)
    data = json.loads(json_output)

    assert 'terms' in data
    assert 'functions' in data
    assert 'hierarchy' in data
    assert 'meta' in data
    assert data['terms'] == [
        {
            'name': 'MyType',
            'label': 'A some label',
            'description': 'A sample type',
            'attributes': {},
        }
    ]
    assert data['functions'] == [
        {
            'name': 'MyFunction',
            'label': 'Function',
            'input_types': [
                {
                    'name': 'int',
                    'label': '',
                }
            ],
            'output_type': {
                'name': 'bool',
                'label': '',
            },
            'attributes': {},
        }
    ]

    assert data['hierarchy'] == [
        {
            'name': None,
            'parent': 'int',
            'relationship': 'dependence',
            'children': ['bool'],
            'attributes': {},
        }
    ]
    assert data['meta'] == {
        'version': '1.0',
        'title': 'SampleOntology',
        'author': 'Author',
        'description': 'Sample description',
        'type': 'Базовый',
        'date': '2024-01-01',
    }


def test_serialize_empty_ontology(serializer):
    ontology = Ontology()
    json_output = serializer.serialize(ontology)
    data = json.loads(json_output)
    assert data == {
        'meta': {
            'version': None,
            'title': None,
            'author': None,
            'description': None,
            'type': None,
            'date': None,
        },
        'terms': [],
        'functions': [],
        'hierarchy': [],
        'figures': [],
    }


def test_serialize_term(serializer):
    term = Term(name='TestTerm', description='A test term', label='A test label')
    serialized_term = serializer._serialize_term(term)
    assert serialized_term == {
        'name': 'TestTerm',
        'label': 'A test label',
        'description': 'A test term',
        'attributes': {},
    }


def test_serialize_function(serializer):
    func = Function(
        name='TestFunction',
        input_types=[FunctionArgument(Term('int'))],
        output_type=FunctionArgument(Term('bool')),
        label='A test function',
    )
    serialized_function = serializer._serialize_function(func)
    assert serialized_function == {
        'name': 'TestFunction',
        'label': 'A test function',
        'input_types': [
            {
                'name': 'int',
                'label': '',
            }
        ],
        'output_type': {
            'name': 'bool',
            'label': '',
        },
        'attributes': {},
    }


def test_serialize_meta(serializer):
    meta = Meta(
        version='1.0',
        title='TestOntology',
        author='Author',
        description='A test ontology',
        type='Базовый',
        date='2024-01-01',
    )
    serialized_meta = serializer._serialize_meta(meta)
    assert serialized_meta == {
        'version': '1.0',
        'title': 'TestOntology',
        'author': 'Author',
        'description': 'A test ontology',
        'type': 'Базовый',
        'date': '2024-01-01',
    }


def test_serialize_ontology_without_meta(serializer):
    ontology = Ontology()
    ontology.add_type(
        Term(name='MyType', description='A sample type', label='A some label')
    )
    ontology.add_function(
        Function(
            name='MyFunction',
            label='Function',
            input_types=[FunctionArgument(Term('int'))],
            output_type=FunctionArgument(Term('bool')),
        )
    )
    ontology.add_relationship(
        Relationship(
            relationship=RelationshipType.DEPENDENCE,
            parent=Term('int'),
            children=[Term('bool')],
        )
    )

    json_output = serializer.serialize(ontology)
    data = json.loads(json_output)

    assert 'terms' in data
    assert 'functions' in data
    assert 'hierarchy' in data
    assert data['meta'] == {
        'version': None,
        'title': None,
        'author': None,
        'description': None,
        'type': None,
        'date': None,
    }
    print(data)
    assert data['terms'] == [
        {
            'name': 'MyType',
            'label': 'A some label',
            'description': 'A sample type',
            'attributes': {},
        }
    ]
    assert data['functions'] == [
        {
            'name': 'MyFunction',
            'label': 'Function',
            'input_types': [
                {
                    'name': 'int',
                    'label': '',
                }
            ],
            'output_type': {
                'name': 'bool',
                'label': '',
            },
            'attributes': {},
        }
    ]

    assert data['hierarchy'] == [
        {
            'name': None,
            'parent': 'int',
            'relationship': 'dependence',
            'children': ['bool'],
            'attributes': {},
        }
    ]



from ontol import (
    Parser,
    Term,
    Function,
    Relationship,
    TermAttributes,
    FunctionAttributes,
    RelationshipAttributes,
)

import pytest


@pytest.fixture
def parser():
    return Parser()


def test_parse_empty_file(parser):
    content: str = ''
    ontology, warnings = parser.parse(content, 'test.ontol')
    assert ontology.meta.version is None
    assert ontology.meta.title is None
    assert ontology.meta.author is None
    assert ontology.meta.description is None
    assert ontology.meta.type is None
    assert ontology.meta.date is not None
    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0
    assert len(warnings) == 0


def test_parse_commented_file(parser):
    content: str = """
    # version: '1.0'
    # title: 'Basic linear algebra'
    # author: 'Firstname Lastname'
    # description: 'Matrices'

    # types:
    # number: 'Number', ''
    # matrix: 'Matrix', ''

    # functions:
    # transpose: 'Transpose' (matrix: '') -> matrix: 'Transposed matrix'

    # hierarchy:
    # matrix composition number
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert ontology.meta.version is None
    assert ontology.meta.title is None
    assert ontology.meta.author is None
    assert ontology.meta.description is None
    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0
    assert len(warnings) == 0


def test_parse_meta_tags(parser):
    content: str = """
    version: '2.5.5'
    title: 'Biology'
    author: 'Firstname Lastname'
    description: 'Biology for high school'
    date: '02.02.2025'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0

    assert ontology.meta is not None

    assert ontology.meta.version == '2.5.5'
    assert ontology.meta.title == 'Biology'
    assert ontology.meta.author == 'Firstname Lastname'
    assert ontology.meta.description == 'Biology for high school'
    assert ontology.meta.date == '02.02.2025'


def test_parse_empty_meta_tags(parser):
    content: str = """
    version: ''
    title: ''
    author: ''
    description: ''
    date: ''
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0

    assert ontology.meta is not None

    assert ontology.meta.version == ''
    assert ontology.meta.title == ''
    assert ontology.meta.author == ''
    assert ontology.meta.description == ''
    assert ontology.meta.date != ''


def test_parse_file_without_meta_tags(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {color: '#ffffff'}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0

    assert ontology.meta is not None

    assert ontology.meta.version is None
    assert ontology.meta.title is None
    assert ontology.meta.author is None
    assert ontology.meta.description is None
    assert ontology.meta.date is not None


def test_parse_file_with_incorrect_meta_tags(parser):
    content: str = """
    university: 'Peter the Great St.Petersburg Polytechnic University'
    """
    with pytest.raises(ValueError):
        parser.parse(content, 'test.ontol')

    content: str = """
    version: 1.0.0
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')
        parser.parse(content, 'test.ontol')

    content: str = """
    version '1.0.0'
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    'version': '1.0.0'
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    'version' '1.0.0'
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_type(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {color: '#ffffff'}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == TermAttributes(color='#ffffff')


def test_parse_type_Wth_multiline_attributes(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {
        color: '#ffffff'
    }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == TermAttributes(color='#ffffff')


def test_parse_type_Wth_multiline_attributes_with_trailing_comma(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {
        color: '#ffffff',
    }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == TermAttributes(color='#ffffff')


def test_parse_type_without_attributes(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == TermAttributes()
    assert len(warnings) == 0


def test_parse_type_with_empty_attributes(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == TermAttributes()
    assert len(warnings) == 0


def test_parse_type_with_incorrect_attributes(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { foo }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { foo, bar, baz }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { foo bar }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { 123 }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов' { color: '#red' }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { 'color': '#red' }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', { color: red }
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_type_with_empty_label_and_desc(parser):
    content: str = """
    types:
    set: '', ''
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == ''
    assert term.description == ''
    assert term.attributes == TermAttributes()
    assert len(warnings) == 2


def test_parse_type_without_label_and_desc(parser):
    content: str = """
    types:
    set: ''
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function(parser):
    content = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов'
    functions:
    descartes: 'Cartesian product' (set: 'First set', set: 'Second set') -> set: 'Result set', { color: '#fff' }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'descartes'
    assert func.label == 'Cartesian product'
    assert func.input_types[0].term.name == 'set'
    assert func.input_types[0].term.label == 'Множество'
    assert func.input_types[0].term.description == 'Коллекция уникальных элементов'
    assert func.input_types[0].label == 'First set'
    assert func.input_types[1].term.name == 'set'
    assert func.input_types[1].term.label == 'Множество'
    assert func.input_types[1].term.description == 'Коллекция уникальных элементов'
    assert func.input_types[1].label == 'Second set'
    assert func.output_type.term.name == 'set'
    assert func.output_type.term.label == 'Множество'
    assert func.output_type.term.description == 'Коллекция уникальных элементов'
    assert func.output_type.label == 'Result set'
    assert func.attributes == FunctionAttributes(color='#fff')
    assert len(warnings) == 0


def test_parse_function_with_empty_attributes(parser):
    content = """
    types:
    number: 'Число', 'Число'
    functions:
    divide: 'Divide two numbers' (number: 'Dividend', number: 'Divisor') -> number: 'Quotient', {}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'divide'
    assert func.label == 'Divide two numbers'
    assert func.input_types[0].term.name == 'number'
    assert func.input_types[0].term.label == 'Число'
    assert func.input_types[0].term.description == 'Число'
    assert func.input_types[0].label == 'Dividend'
    assert func.input_types[1].term.name == 'number'
    assert func.input_types[1].term.label == 'Число'
    assert func.input_types[1].term.description == 'Число'
    assert func.input_types[1].label == 'Divisor'
    assert func.output_type.term.name == 'number'
    assert func.output_type.term.label == 'Число'
    assert func.output_type.term.description == 'Число'
    assert func.output_type.label == 'Quotient'
    assert func.attributes == FunctionAttributes()
    assert len(warnings) == 0


def test_parse_function_with_missing_output(parser):
    content = """
    types:
    number: '', ''
    functions:
    add: 'Addition' (number: 'First number', number: 'Second number') -> , {color: '#fff'}
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function_with_missing_label(parser):
    content = """
    functions:
    add(number: 'First number', number: 'Second number') -> number: 'Result', {color: '#fff'}
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function_without_arguments(parser):
    content = """
    types:
    time: 'Current time', 'Current time'
    functions:
    now: 'Returns current time' () -> time: 'Current time'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'now'
    assert func.label == 'Returns current time'
    assert func.input_types == []
    assert func.output_type.term.name == 'time'
    assert func.output_type.label == 'Current time'
    assert len(warnings) == 0


def test_parse_with_empty_labels(parser):
    content = """
    types:
    set: '', ''
    functions:
    descartes: '' (set: '', set: '') -> set: ''
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'descartes'
    assert func.label == ''
    assert func.input_types[0].term.name == 'set'
    assert func.input_types[0].label == ''
    assert func.input_types[1].term.name == 'set'
    assert func.input_types[1].label == ''
    assert func.output_type.term.name == 'set'
    assert func.output_type.label == ''
    print('\n\n'.join(warnings))
    assert len(warnings) == 6


def test_parse_heierarchy(parser):
    content = """
    types:
    set: '', ''
    element: '', ''
    hierarchy:
    element inheritance set
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.hierarchy) == 1

    rel: Relationship = ontology.hierarchy[0]
    assert rel.parent.name == 'element'
    assert rel.relationship.value == 'inheritance'
    assert rel.children[0].name == 'set'
    assert rel.attributes == RelationshipAttributes()
    assert len(warnings) == 4


def test_parse_heierarchy_with_attributes(parser):
    content = """
    types:
    set: '', ''
    element: '', ''
    hierarchy:
    element inheritance set, { color: '#red' }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.hierarchy) == 1

    rel: Relationship = ontology.hierarchy[0]
    assert rel.parent.name == 'element'
    assert rel.relationship.value == 'inheritance'
    assert rel.children[0].name == 'set'
    assert rel.attributes == RelationshipAttributes(color='#red')
    assert len(warnings) == 4


def test_parse_heierarchy_with_multiline_attributes(parser):
    content = """
    types:
    set: '', ''
    element: '', ''
    hierarchy:
    element inheritance set, {
        color: '#red'
    }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.hierarchy) == 1

    rel: Relationship = ontology.hierarchy[0]
    assert rel.parent.name == 'element'
    assert rel.relationship.value == 'inheritance'
    assert rel.children[0].name == 'set'
    assert rel.attributes == RelationshipAttributes(color='#red')
    assert len(warnings) == 4


def test_parse_heierarchy_with_empty_attributes(parser):
    content = """
    types:
    set: '', ''
    element: '', ''
    hierarchy:
    element inheritance set, {  }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.hierarchy) == 1

    rel: Relationship = ontology.hierarchy[0]
    assert rel.parent.name == 'element'
    assert rel.relationship.value == 'inheritance'
    assert len(rel.children) == 1
    assert rel.children[0].name == 'set'
    assert rel.attributes == RelationshipAttributes()
    assert len(warnings) == 4


def test_parse_heierarchy_unexpected_relationship(parser):
    content = """
    types:
    set: '', ''
    element: '', ''
    hierarchy:
    element loves set, {  }
    """
    with pytest.raises(ValueError):
        parser.parse(content, 'test.ontol')


def test_parse_heierarchy_with_undefined_types(parser):
    content = """
    types:
    set: '', ''
    hierarchy:
    element composition set
    """
    with pytest.raises(ValueError):
        parser.parse(content, 'test.ontol')

    content = """
    types:
    set: '', ''
    hierarchy:
    set composition element
    """
    with pytest.raises(ValueError):
        parser.parse(content, 'test.ontol')

    content = """
    types:
    set: '', ''
    hierarchy:
    element composition element
    """
    with pytest.raises(ValueError):
        parser.parse(content, 'test.ontol')


def test_combined_parsing(parser):
    content = """
    # Ontology information
    version: '1.0'
    title: 'Basic linear algebra'

    # Types we have
    types:
    # todo: Add laso complex numbers
    number: 'Number', 'Some real number from R field'

    # Other type
    types:
    matrix: 'Matrix', 'Rectangular array or table of numbers'

    # Functions we have
    functions:
    # Describe more functions
    transpose: 'Transpose' (matrix: 'Initial matrix') -> matrix: 'Transposed matrix'

    # And relations we have
    hierarchy:
    # All also vectors
    matrix composition number
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 2
    assert len(ontology.functions) == 1
    assert len(ontology.hierarchy) == 1

    assert ontology.meta is not None

    assert len(warnings) == 0



import os
from urllib.parse import urlparse
import requests
from sly import Lexer as BaseLexer, Parser as BaseParser
from datetime import datetime
from typing import Literal, Optional, Any, Type
from dataclasses import fields

from ontol import (
    constants,
    Ontology,
    Figure,
    Term,
    Function,
    Relationship,
    Meta,
    FunctionArgument,
    RelationshipType,
    RelationshipAttributes,
    TermAttributes,
    FunctionAttributes,
    RelationshipDirection,
)


class Lexer(BaseLexer):
    tokens = {
        TYPES_BLOCK,
        FUNCTIONS_BLOCK,
        HIERARCHY_BLOCK,
        FIGURE_BLOCK,
        IMPORT_KEYWORD,
        FROM_KEYWORD,
        AS_KEYWORD,
        STRING,
        IDENTIFIER,
        LBRACE,
        RBRACE,
        COLON,
        COMMA,
        ASTERISK,
        ARROW,
        LPAREN,
        RPAREN,
        NEWLINE,
    }

    # String containing ignored characters (spaces and tabs)
    ignore: str = ' \t'

    # Regular expression rules for tokens
    TYPES_BLOCK: str = r'\btypes\b'
    FUNCTIONS_BLOCK: str = r'\bfunctions\b'
    HIERARCHY_BLOCK: str = r'\bhierarchy\b'
    FIGURE_BLOCK: str = r'\bfigure\b'

    IMPORT_KEYWORD: str = r'\bimport\b'
    FROM_KEYWORD: str = r'\bfrom\b'
    AS_KEYWORD: str = r'\bas\b'

    STRING: str = r'\'[^\']*\'|\"[^\"]*\"'
    IDENTIFIER: str = r'[a-zA-Z_][a-zA-Z0-9_]*'

    LBRACE: str = r'\{'
    RBRACE: str = r'\}'
    COLON: str = r':'
    COMMA: str = r','
    ASTERISK: str = r'\*'
    ARROW: str = r'->'

    LPAREN: str = r'\('
    RPAREN: str = r'\)'

    ignore_comment: str = r'\#.*'

    # Newline handling
    @_(r'\n+')
    def NEWLINE(self, t):
        self.lineno += t.value.count('\n')
        return t

    # Strings handling
    def STRING(self, t):
        t.value = t.value[1:-1]  # Remove quotes
        return t

    def error(self, t) -> None:
        self.index += 1
        raise SyntaxError(f"{constants.error_prefix} illegal character '{t.value[0]}'")


class Parser(BaseParser):
    tokens = Lexer.tokens
    expected_shift_reduce: int = 26

    def __init__(self) -> None:
        self.__ontology: Ontology = Ontology()
        self.__warnings: list[str] = []

    def parse(self, file_content: str, file_path: str) -> tuple[Ontology, list[str]]:
        self.__warnings.clear()

        # FIX: fix EOF issue
        file_content += '\n'

        self.__lines: list[str] = file_content.splitlines()
        self.__file_path: str = file_path
        self.__ontology: Ontology = Ontology()

        lexer: Lexer = Lexer()
        tokens: list = lexer.tokenize(file_content)

        super().parse(tokens)

        return self.__ontology, self.__warnings

    def _get_exception_message(
        self, token, message: str, type: Literal['warning', 'error'] = 'warning'
    ) -> str:
        line_number = token.lineno
        line: str = self.__lines[line_number - 1]
        line_padding: int = 4
        message_prefix: str = (
            constants.warning_prefix if type == 'warning' else constants.error_prefix
        )

        final_message: str = f'File "{self.__file_path}", line {line_number}'
        final_message += f'\n{" " * line_padding}{line}'
        if token is not None:
            line_start_index: int = sum(
                len(line) + 1 for line in self.__lines[: line_number - 1]
            )
            column_index: int = token.index - line_start_index
            final_message += f'\n{" " * line_padding}{" " * column_index}^'
        final_message += f'\n{message_prefix} {message[0].lower() + message[1:]}'

        return final_message

    def _add_warning(self, token, message: str) -> None:
        final_message: str = self._get_exception_message(token, message, 'warning')
        self.__warnings.append(final_message)

    def _tokenized_attributes_to_dict(
        self,
        tokenized_attributes: list[tuple],
        attributesClass: Type[TermAttributes]
        | Type[FunctionAttributes]
        | Type[RelationshipAttributes],
    ) -> dict[str, Any]:
        allowed_attributes: list[str] = [
            field.name for field in fields(attributesClass)
        ]
        attributes: dict[str, str] = {}

        for attribute_pair in tokenized_attributes:
            if attribute_pair[0].value not in allowed_attributes:
                raise ValueError(
                    self._get_exception_message(
                        attribute_pair[0],
                        f'Unexpected attribute {attribute_pair[0].value}. One of the following was expected: {", ".join(allowed_attributes)}',
                        'error',
                    )
                )
            attributes[attribute_pair[0].value] = attribute_pair[1].value

        return attributes

    @_('statement_list')
    def program(self, p) -> Ontology:
        if not self.__ontology.meta.date:
            self.__ontology.meta.date = datetime.today().strftime('%Y-%m-%d')
        return self.__ontology

    @_('statement_list statement', '')
    def statement_list(self, p) -> None:
        pass

    @_('statement')
    def statement_list(self, p) -> None:
        pass

    @_('IDENTIFIER COLON STRING NEWLINE')
    def statement(self, p) -> None:
        allowed_meta_tags = [field.name for field in fields(Meta)]

        if p.IDENTIFIER not in allowed_meta_tags:
            raise ValueError(
                self._get_exception_message(
                    p._slice[0],
                    f'Unexpected meta tag. One of the following was expected: {", ".join(allowed_meta_tags)}',
                    'error',
                )
            )

        if not p.STRING:
            self._add_warning(p._slice[1], 'Version value is empty')

        setattr(self.__ontology.meta, p.IDENTIFIER, p.STRING)

    @_('IDENTIFIER')
    def import_identifier(self, p) -> tuple:
        return (p._slice[0], None)

    @_('IDENTIFIER AS_KEYWORD IDENTIFIER')
    def import_identifier(self, p) -> tuple:
        return (p._slice[0], p._slice[2])

    @_(
        'import_identifiers_list COMMA import_identifier',
        'import_identifiers_list COMMA NEWLINE import_identifier',
    )
    def import_identifiers_list(self, p) -> list:
        return p.import_identifiers_list + [p.import_identifier]

    @_('import_identifier')
    def import_identifiers_list(self, p) -> list:
        return [p.import_identifier]

    @_('')
    def import_identifiers_list(self, p) -> list:
        return []

    @_(
        'LBRACE import_identifiers_list RBRACE',
        'LBRACE NEWLINE import_identifiers_list RBRACE',
        'LBRACE import_identifiers_list NEWLINE RBRACE',
        'LBRACE NEWLINE import_identifiers_list NEWLINE RBRACE',
        'LBRACE NEWLINE import_identifiers_list COMMA NEWLINE RBRACE',
    )
    def imported_identifiers(self, p) -> list[tuple]:
        return p.import_identifiers_list

    @_('')
    def imported_identifiers(self, p) -> list[tuple]:
        return []

    def _validate_src(self, src: str) -> bool:
        try:
            result = urlparse(src)
            return bool(result.scheme and result.netloc)
        except AttributeError:
            return False

    def _add_definition_if_does_not_exist(
        self, definition: Term | Function | Relationship
    ) -> None:
        if definition.name is not None:
            existing_definition: Optional[Term | Function | Relationship] = (
                self.__ontology.find_definition_by_name(definition.name)
            )
            if existing_definition is not None:
                return

        if isinstance(definition, Term):
            self.__ontology.add_type(definition)
        elif isinstance(definition, Function):
            self.__ontology.add_function(definition)
        elif isinstance(definition, Relationship):
            self.__ontology.add_relationship(definition)

    def _import_ontology(
        self,
        src_token,
        import_tokens: Optional[list[tuple[Any, Any]]] = None,
        asterisk_token=None,
    ) -> None:
        content: str = ''
        file_path = src_token.value

        if self._validate_src(file_path):
            try:
                response = requests.get(file_path)
                if isinstance(response.content, bytes):
                    content = response.content.decode('utf-8')
                else:
                    content = response.text
            except Exception as error:
                raise ValueError(
                    self._get_exception_message(
                        src_token,
                        f'Could not fetch the file: {error}',
                        'error',
                    )
                )
        else:
            if not os.path.isabs(file_path):
                base_dir = os.path.dirname(self.__file_path)
                file_path = os.path.join(base_dir, file_path)
            try:
                with open(file_path, 'r', encoding='utf-8') as file:
                    content = file.read()
            except FileNotFoundError:
                raise ValueError(
                    self._get_exception_message(
                        src_token,
                        f"The file '{file_path}' was not found",
                        'error',
                    )
                )
            except PermissionError:
                raise ValueError(
                    self._get_exception_message(
                        src_token,
                        f"Permission denied when trying to read the file '{file_path}'.",
                        'error',
                    )
                )
            except Exception as error:
                raise ValueError(
                    self._get_exception_message(
                        src_token,
                        f'Could not read the file: {str(error)}',
                        'error',
                    )
                )

        parser: Parser = Parser()
        ontology, warnings = parser.parse(content, file_path)
        self.__warnings.extend(warnings)

        if import_tokens is not None:
            for name_token, alias_token in import_tokens:
                definition: Optional[Term | Function | Relationship] = (
                    ontology.find_definition_by_name(name_token.value)
                )
                if definition is None:
                    raise ValueError(
                        self._get_exception_message(
                            name_token,
                            f'Could not import definition {name_token.value}',
                            'error',
                        )
                    )

        definitions: list[Term | Function | Relationship] = (
            ontology.types + ontology.functions + ontology.hierarchy
        )
        for definition in definitions:
            if definition.name is None:
                continue

            name_token, alias_token = None, None
            if import_tokens is not None:
                name_token, alias_token = next(
                    (
                        (name_token, alias_token)
                        for name_token, alias_token in import_tokens
                        if definition.name == name_token.value
                    ),
                    (None, None),
                )
                if name_token is None:
                    continue

            if alias_token is not None:
                definition.name = alias_token.value

            exception_token: Any = None
            if alias_token is not None:
                exception_token = alias_token
            elif name_token is not None:
                exception_token = name_token
            else:
                exception_token = asterisk_token

            existing_definition: Optional[Term | Function | Relationship] = (
                self.__ontology.find_definition_by_name(definition.name)
            )
            if existing_definition is not None:
                raise ValueError(
                    self._get_exception_message(
                        exception_token,
                        f'Definition {definition.name} has already been declared',
                        'error',
                    )
                )

            if isinstance(definition, Term):
                self.__ontology.add_type(definition)
            if isinstance(definition, Function):
                self.__ontology.add_function(definition)
                for arg in definition.input_types:
                    self._add_definition_if_does_not_exist(arg.term)
                self._add_definition_if_does_not_exist(definition.output_type.term)
            elif isinstance(definition, Relationship):
                self.__ontology.add_relationship(definition)
                self._add_definition_if_does_not_exist(definition.parent)
                self._add_definition_if_does_not_exist(definition.children[0])

    @_('IMPORT_KEYWORD imported_identifiers FROM_KEYWORD STRING')
    def statement(self, p) -> None:
        self._import_ontology(p._slice[3], p.imported_identifiers)

    @_('IMPORT_KEYWORD ASTERISK FROM_KEYWORD STRING')
    def statement(self, p) -> None:
        self._import_ontology(p._slice[3], None, p._slice[1])

    @_('TYPES_BLOCK COLON NEWLINE type_list')
    def statement(self, p) -> None:
        pass

    @_(
        'type_list type NEWLINE',
        'NEWLINE type_list',
        '',
    )
    def type_list(self, p) -> None:
        pass

    @_('IDENTIFIER COLON STRING COMMA STRING attributes')
    def type(self, p) -> None:
        existing_definition: Optional[Term | Function | Relationship] = (
            self.__ontology.find_definition_by_name(p.IDENTIFIER)
        )

        if existing_definition is not None:
            raise ValueError(
                self._get_exception_message(
                    p._slice[0],
                    f'Definition {p.IDENTIFIER} has already been declared',
                    'error',
                )
            )

        attributes: dict[str, Any] = self._tokenized_attributes_to_dict(
            p.attributes, TermAttributes
        )

        term = Term(
            name=p.IDENTIFIER,
            label=p.STRING0,
            description=p.STRING1,
            attributes=TermAttributes(**attributes),
        )

        if not p.STRING0:
            self._add_warning(p._slice[2], 'Term label is empty')

        if not p.STRING1:
            self._add_warning(p._slice[4], 'Term description is empty')

        self.__ontology.add_type(term)

    @_('FUNCTIONS_BLOCK COLON NEWLINE function_list')
    def statement(self, p) -> None:
        pass

    @_(
        'function_list function NEWLINE',
        'NEWLINE function_list',
        '',
    )
    def function_list(self, p) -> None:
        pass

    def _tokenized_function_attributes_to_dict(
        self,
        tokenized_attributes: list[tuple],
    ) -> dict[str, Any]:
        attributes: dict[str, Any] = self._tokenized_attributes_to_dict(
            tokenized_attributes, FunctionAttributes
        )

        if 'type' in attributes:
            attributes['type'] = RelationshipType.from_str(attributes['type'])
            if attributes['type'] is None:
                for key_token, value_token in tokenized_attributes:
                    if key_token.value == 'type':
                        raise ValueError(
                            self._get_exception_message(
                                value_token,
                                f'Unexpected type type. One of the following was expected: {", ".join(member.value for member in RelationshipType)}',
                                'error',
                            )
                        )

        return attributes

    @_('IDENTIFIER COLON STRING params ARROW IDENTIFIER COLON STRING attributes')
    def function(self, p) -> None:
        existing_definition: Optional[Term | Function | Relationship] = (
            self.__ontology.find_definition_by_name(p.IDENTIFIER0)
        )

        if existing_definition is not None:
            raise ValueError(
                self._get_exception_message(
                    p._slice[0],
                    f'Definition {p.IDENTIFIER0} has already been declared',
                    'error',
                )
            )

        input_types = p.params

        output_term: Optional[Term] = self.__ontology.find_term_by_name(p.IDENTIFIER1)

        if output_term is None:
            raise ValueError(
                self._get_exception_message(
                    p._slice[5], f'Undefined term {p.IDENTIFIER1}', 'error'
                )
            )

        output_type: FunctionArgument = FunctionArgument(output_term, p.STRING1)

        attributes: dict[str, Any] = self._tokenized_function_attributes_to_dict(
            p.attributes
        )

        function: Function = Function(
            name=p.IDENTIFIER0,
            label=p.STRING0,
            input_types=input_types,
            output_type=output_type,
            attributes=FunctionAttributes(**attributes),
        )

        if not p.STRING0:
            self._add_warning(p._slice[2], 'Label is empty')

        if not p.STRING1:
            self._add_warning(p._slice[7], 'Output term label is empty')

        self.__ontology.add_function(function)

    @_(
        'LPAREN param_list RPAREN',
        'LPAREN NEWLINE param_list RPAREN',
        'LPAREN param_list NEWLINE RPAREN',
        'LPAREN NEWLINE param_list NEWLINE RPAREN',
        'LPAREN NEWLINE param_list COMMA NEWLINE RPAREN',
    )
    def params(self, p) -> list[FunctionArgument]:
        params: list[FunctionArgument] = []

        for param in p.param_list:
            term_token = param[0]
            term_name: str = term_token.value
            label_token = param[1]
            param_label: str = label_token.value

            term: Optional[Term] = self.__ontology.find_term_by_name(term_name)

            if term is None:
                raise ValueError(
                    self._get_exception_message(
                        term_token, f'Undefined term {term_name}', 'error'
                    )
                )

            if not param_label:
                self._add_warning(label_token, 'Parameter label is empty')

            params.append(FunctionArgument(term, param_label))

        return params

    @_('')
    def param_list(self, p) -> list[tuple]:
        return []

    @_('param')
    def param_list(self, p) -> list[tuple]:
        return [p.param]

    @_(
        'param_list COMMA param',
        'param_list COMMA NEWLINE param',
    )
    def param_list(self, p) -> list[tuple]:
        return p.param_list + [p.param]

    @_('IDENTIFIER COLON STRING')
    def param(self, p) -> tuple:
        return (p._slice[0], p._slice[2])

    @_('HIERARCHY_BLOCK COLON NEWLINE hierarchy_list')
    def statement(self, p) -> None:
        pass

    @_(
        'hierarchy_list relationship NEWLINE',
        'NEWLINE hierarchy_list',
        '',
    )
    def hierarchy_list(self, p) -> None:
        pass

    def _tokenized_relationship_attributes_to_dict(
        self,
        tokenized_attributes: list[tuple],
    ) -> dict[str, Any]:
        attributes: dict[str, Any] = self._tokenized_attributes_to_dict(
            tokenized_attributes, RelationshipAttributes
        )

        if 'direction' in attributes:
            attributes['direction'] = RelationshipDirection.from_str(
                attributes['direction']
            )
            if attributes['direction'] is None:
                for key_token, value_token in tokenized_attributes:
                    if key_token.value == 'direction':
                        raise ValueError(
                            self._get_exception_message(
                                value_token,
                                f'Unexpected direction type. One of the following was expected: {", ".join(member.value for member in RelationshipDirection)}',
                                'error',
                            )
                        )

        return attributes

    def _add_relationship(
        self,
        name_token,
        parent_token,
        relationship_type_token,
        child_token,
        attributes_tokens,
    ) -> None:
        if name_token is not None:
            existing_definition: Optional[Term | Function | Relationship] = (
                self.__ontology.find_definition_by_name(name_token.value)
            )
            if existing_definition is not None:
                raise ValueError(
                    self._get_exception_message(
                        name_token,
                        f'Definition {name_token.value} has already been declared',
                        'error',
                    )
                )

        parent: Optional[Term] = self.__ontology.find_term_by_name(parent_token.value)

        if parent is None:
            raise ValueError(
                self._get_exception_message(
                    parent_token, f'Undefined term {parent_token.value}', 'error'
                )
            )

        relationship_type: Optional[RelationshipType] = RelationshipType.from_str(
            relationship_type_token.value
        )

        if relationship_type is None:
            raise ValueError(
                self._get_exception_message(
                    relationship_type_token,
                    f'Unexpected relationship type. One of the following was expected: {", ".join(member.value for member in RelationshipType)}',
                    'error',
                )
            )

        child: Optional[Term] = self.__ontology.find_term_by_name(child_token.value)

        if child is None:
            raise ValueError(
                self._get_exception_message(
                    child_token, f'Undefined term {child_token.value}', 'error'
                )
            )

        children: list[Term] = [child]

        attributes: dict[str, Any] = self._tokenized_relationship_attributes_to_dict(
            attributes_tokens
        )

        relationship: Relationship = Relationship(
            name=name_token.value if name_token else None,
            parent=parent,
            relationship=relationship_type,
            children=children,
            attributes=RelationshipAttributes(**attributes),
        )
        self.__ontology.add_relationship(relationship)

    @_('IDENTIFIER IDENTIFIER IDENTIFIER attributes')
    def relationship(self, p) -> None:
        self._add_relationship(
            None,
            p._slice[0],
            p._slice[1],
            p._slice[2],
            p.attributes,
        )

    @_('IDENTIFIER COLON IDENTIFIER IDENTIFIER IDENTIFIER attributes')
    def relationship(self, p) -> None:
        self._add_relationship(
            p._slice[0],
            p._slice[2],
            p._slice[3],
            p._slice[4],
            p.attributes,
        )

    @_('FIGURE_BLOCK STRING COLON NEWLINE figure_list')
    def statement(self, p) -> None:
        figure: Figure = Figure(name=p.STRING)

        for token in p.figure_list:
            definition: Optional[Term | Function | Relationship] = (
                self.__ontology.find_definition_by_name(token.value)
            )
            if definition is None:
                raise ValueError(
                    self._get_exception_message(
                        token,
                        f'Undefined identifier {token.value}',
                        'error',
                    )
                )
            if isinstance(definition, Term):
                figure.types.append(definition)
            elif isinstance(definition, Function):
                figure.functions.append(definition)
            elif isinstance(definition, Relationship):
                figure.hierarchy.append(definition)

        self.__ontology.add_figure(figure)

    @_('figure_list IDENTIFIER NEWLINE')
    def figure_list(self, p) -> list[Any]:
        return p.figure_list + [p._slice[1]]

    @_(
        'NEWLINE figure_list',
    )
    def figure_list(self, p) -> list[Any]:
        return p.figure_list

    @_('')
    def figure_list(self, p) -> list[Any]:
        return []

    @_(
        'COMMA LBRACE attribute_list RBRACE',
        'COMMA LBRACE NEWLINE attribute_list RBRACE',
        'COMMA LBRACE attribute_list NEWLINE RBRACE',
        'COMMA LBRACE NEWLINE attribute_list NEWLINE RBRACE',
        'COMMA LBRACE NEWLINE attribute_list COMMA NEWLINE RBRACE',
    )
    def attributes(self, p) -> list[tuple[Any, Any]]:
        return p.attribute_list

    @_('')
    def attributes(self, p) -> list[tuple[Any, Any]]:
        return []

    @_(
        'attribute_list COMMA attribute',
        'attribute_list COMMA NEWLINE attribute',
    )
    def attribute_list(self, p) -> list[tuple[Any, Any]]:
        return p.attribute_list + [p.attribute]

    @_('attribute')
    def attribute_list(self, p) -> list[tuple[Any, Any]]:
        return [p.attribute]

    @_('')
    def attribute_list(self, p) -> list[tuple[Any, Any]]:
        return []

    @_('IDENTIFIER COLON STRING')
    def attribute(self, p) -> tuple[Any, Any]:
        if not p.STRING:
            self._add_warning(p._slice[2], 'Attribute value is empty')

        return (p._slice[0], p._slice[2])

    @_('NEWLINE')
    def statement(self, p) -> None:
        pass

    def error(self, p) -> None:
        if p:
            raise SyntaxError(
                self._get_exception_message(
                    p,
                    f'Syntax error ({p.type})',
                    'error',
                )
            )

        raise SyntaxError(f'{constants.error_prefix} Syntax error at EOF')



import os
import tempfile

from ontol import CLI
from unittest.mock import MagicMock, patch

import pytest


@pytest.fixture
def cli():
    return CLI()


@pytest.fixture
def sample_ontology_file():
    content = """
    version: '1.0'
    """
    with tempfile.NamedTemporaryFile(suffix='.ontol', delete=False) as tmp_file:
        tmp_file.write(content.encode('utf-8'))
        tmp_file_path = tmp_file.name
    return tmp_file_path


def test_parse_file(cli, sample_ontology_file):
    cli.parse_file(sample_ontology_file)

    json_file_path: str = os.path.splitext(sample_ontology_file)[0] + '.json'
    puml_file_path: str = os.path.splitext(sample_ontology_file)[0] + '.puml'

    assert os.path.exists(json_file_path)
    assert os.path.exists(puml_file_path)

    os.remove(sample_ontology_file)
    os.remove(json_file_path)
    os.remove(puml_file_path)


def test_render_plantuml_to_png(cli):
    cli.render_plantuml_to_png = MagicMock()
    puml_file_path = 'test.puml'

    cli.render_plantuml_to_png(puml_file_path)

    cli.render_plantuml_to_png.assert_called_with(puml_file_path)


def test_watch_file(cli):
    with (
        patch('watchdog.observers.Observer.schedule') as mock_schedule,
        patch('watchdog.observers.Observer.start') as mock_start,
        patch('watchdog.observers.Observer.join') as mock_join,
        patch('time.sleep', side_effect=KeyboardInterrupt),
    ):
        mock_start.return_value = None
        mock_join.return_value = None

        cli.watch_file('test.ontol')

        mock_schedule.assert_called()
        mock_start.assert_called()
        mock_join.assert_called()


def test_watch_file_no_changes(cli):
    with (
        patch('watchdog.observers.Observer.schedule') as mock_schedule,
        patch('watchdog.observers.Observer.start') as mock_start,
        patch('watchdog.observers.Observer.join') as mock_join,
        patch('time.sleep', side_effect=KeyboardInterrupt),
    ):
        mock_start.return_value = None
        mock_join.return_value = None

        cli.watch_file('nonexistent_file.ontol')

        mock_schedule.assert_called()
        mock_start.assert_called()
        mock_join.assert_called()



from . import constants
from .oast import (
    Ontology,
    Figure,
    Term,
    Function,
    Relationship,
    Meta,
    FunctionArgument,
    RelationshipType,
    FunctionAttributes,
    RelationshipAttributes,
    RelationshipDirection,
    TermAttributes,
)
from .parser import Parser
from .plantuml import PlantUML
from .serializer import JSONSerializer
from .retranslator import Retranslator
from .ai import AI
from .cli import CLI


__all__ = (
    'constants',
    'Ontology',
    'Figure',
    'Term',
    'Function',
    'Relationship',
    'FunctionAttributes',
    'RelationshipAttributes',
    'RelationshipDirection',
    'TermAttributes',
    'Meta',
    'FunctionArgument',
    'RelationshipType',
    'Parser',
    'PlantUML',
    'JSONSerializer',
    'Retranslator',
    'AI',
    'CLI',
)



import re
import warnings
from typing import Literal, Optional, Any

from pydantic import BaseModel, Field

from langchain_community.llms import Ollama
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.runnables.utils import Input
from langchain_core.output_parsers import PydanticOutputParser
from langchain.output_parsers import RetryOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import RunnableLambda, RunnableParallel

from ontol import (
    Ontology,
    Relationship,
    Function,
    RelationshipAttributes,
    RelationshipDirection,
    RelationshipType,
    Term,
)

warnings.filterwarnings('ignore')


class JsonExtractor(Runnable):
    json_pattern = r'\{.*\}'

    def invoke(
        self, input: Input, config: Optional[RunnableConfig] = None, **kwargs: Any
    ) -> str:
        matches = re.findall(JsonExtractor.json_pattern, input, re.DOTALL)

        if matches:
            return matches[-1].strip().replace('\\\\', '\\')

        return input


class RelationshipSchema(BaseModel):
    parent: str
    child: str
    relationship: Literal[
        'dependence',
        'association',
        'directAssociation',
        'inheritance',
        'implementation',
        'aggregation',
        'composition',
    ]
    label: Optional[str]
    is_bidirectional: bool
    comment: str


class HierarchyGenerationSchema(BaseModel):
    relationships: list[RelationshipSchema] = Field(
        description='Given the UML classes names, labels and functions where they are used, provide possible class diagram relationships. Also provide short comment why chose this relationship.'
    )


parser = PydanticOutputParser(pydantic_object=HierarchyGenerationSchema)

template = """
### Instructions

As a highly skilled UML expert, your task is to thoroughly analyze the classes and any associated functions provided.
Using this information, you will determine all possible relationships in a class diagram. Adhere to the following requirements:

1. Include relationships that define:
   - Parent class
   - Child class
   - Relationship type (e.g., inheritance, composition, aggregation, association, dependency)
   - Possible arrow label in UML style.
   - Additional comments, if necessary
   - Whether the relationship is bidirectional (true or false)
2. There can be no bidirectional inheritance relationships.
3. Derive as many relevant relationships as possible by examining the functions (methods) where these classes may be used.
4. Output only a JSON object following the format below (with each relationship as an item in a JSON array). Do not include any additional text or explanation beyond the JSON object.

### Context

Classes (name, label, and description, comma separated; may be empty):
```
{classes}
```

Functions (name, label, input arguments, and output argument with descriptions; may be empty):
```
{functions}
```

Format Instructions:
```
{format_instructions}
```

Your final output should be a JSON object listing all identified relationships. Return only this JSON object and nothing else.
"""

prompt = PromptTemplate(
    template=template,
    input_variables=['classes', 'functions'],
    partial_variables={'format_instructions': parser.get_format_instructions()},
)


class AI:
    def _format_string(self, string: str) -> str:
        return string.replace('\\n', '').strip().lower()

    def _format_terms(self, ontology: Ontology) -> str:
        return '\n'.join(
            [
                f"'{self._format_string(t.name)}', '{self._format_string(t.label)}', '{self._format_string(t.description)}'"
                for t in ontology.types
            ]
        )

    def _format_function(self, function: Function) -> str:
        function_line = f"'{self._format_string(function.name)}', '{self._format_string(function.label)}', ("
        args = ''
        for i, arg in enumerate(function.input_types):
            args += f"'{self._format_string(arg.term.name)}': '{self._format_string(arg.label)}'"
            if i < len(function.input_types) - 1:
                args += ', '
        function_line += (
            f'{args}) -> '
            f"'{self._format_string(function.output_type.term.name)}': '{self._format_string(function.output_type.label)}'"
        )

        return function_line

    def _format_functions(self, ontology: Ontology) -> str:
        return '\n'.join([self._format_function(f) for f in ontology.functions])

    def does_edge_exist(
        self, hierarchy: list[Relationship], relationship: Relationship
    ):
        for rel in hierarchy:
            if (
                rel.parent == relationship.parent
                and rel.children[0] == relationship.children[0]
            ):
                return True
        return False

    def generate_hierarchy(
        self, ontology: Ontology, model: str, temperature: float = 0.0
    ) -> tuple[list[Relationship], list[str]]:
        llm = Ollama(model=model, temperature=temperature)
        chain = prompt | llm | JsonExtractor() | parser

        relationships: list[Relationship] = []
        comments: list[str] = []
        formatted_terms = self._format_terms(ontology)
        formatted_functions = self._format_functions(ontology)

        try:
            response = chain.invoke(
                {'classes': formatted_terms, 'functions': formatted_functions}
            )

            for relationship in response.relationships:
                if relationship.relationship == 'inheritance':
                    relationship.parent, relationship.child = (
                        relationship.child,
                        relationship.parent,
                    )

                parent: Optional[Term] = ontology.find_term_by_name(relationship.parent)
                if parent is None:
                    continue

                relationship_type: Optional[RelationshipType] = (
                    RelationshipType.from_str(relationship.relationship)
                )
                if relationship_type is None:
                    continue

                child: Optional[Term] = ontology.find_term_by_name(relationship.child)
                if child is None:
                    continue

                children: list[Term] = [child]

                attributes: RelationshipAttributes = RelationshipAttributes(
                    title=relationship.label
                )
                if relationship.is_bidirectional:
                    attributes.direction = RelationshipDirection.BIDIRECTIONAL

                rel = Relationship(
                    parent=parent,
                    relationship=relationship_type,
                    children=children,
                    attributes=attributes,
                )

                if self.does_edge_exist(ontology.hierarchy + relationships, rel):
                    continue

                relationships.append(rel)
                comments.append(relationship.comment)
        except Exception as e:
            print(e)

        return relationships, comments



import copy
from enum import Enum
from typing import Optional
from dataclasses import dataclass, field


@dataclass
class Meta:
    version: Optional[str] = None
    title: Optional[str] = None
    author: Optional[str] = None
    description: Optional[str] = None
    type: Optional[str] = None
    date: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f'Meta(version={self.version}, title={self.title}, author={self.author}, '
            f'description={self.description}, type={self.type}, '
            f'date={self.date})'
        )

    def with_new_name(self, new_name: str) -> 'Meta':
        return Meta(
            version=self.version,
            title=new_name,
            author=self.author,
            description=self.description,
            type=self.type,
            date=self.date,
        )


@dataclass
class TermAttributes:
    color: Optional[str] = None
    note: Optional[str] = None

    def __repr__(self) -> str:
        return f'TermAttributes(color={self.color}, note={self.note})'


@dataclass
class Term:
    name: str
    label: str = ''
    description: str = ''
    attributes: TermAttributes = field(default_factory=TermAttributes)

    def __repr__(self) -> str:
        return f'Term(name={self.name}, label={self.label}, description={self.description}, attributes={self.attributes})'


class RelationshipDirection(Enum):
    FORWARD = 'forward'
    BACKWARD = 'backward'
    BIDIRECTIONAL = 'bidirectional'

    @classmethod
    def from_str(cls, value: str):
        return cls._value2member_map_.get(value, None)

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in cls._value2member_map_


class RelationshipType(Enum):
    # TODO: change name by Novikov opinion
    DEPENDENCE = 'dependence'
    ASSOCIATION = 'association'
    DIRECT_ASSOCIATION = 'directAssociation'
    INHERITANCE = 'inheritance'
    IMPLEMENTATION = 'implementation'
    AGGREGATION = 'aggregation'
    COMPOSITION = 'composition'

    @classmethod
    def from_str(cls, value: str) -> Optional['RelationshipType']:
        return cls._value2member_map_.get(value, None)

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in cls._value2member_map_


@dataclass
class FunctionAttributes:
    color: Optional[str] = None
    colorArrow: Optional[str] = None
    type: Optional[RelationshipType] = None
    inputTitle: Optional[str] = None
    outputTitle: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f'FunctionAttributes(color={self.color}, '
            f'colorArrow={self.colorArrow}, '
            f'type={self.type.value if self.type else self.type}, '
            f'inputTitle={self.inputTitle}, '
            f'outputTitle={self.outputTitle})'
        )


@dataclass
class FunctionArgument:
    term: Term
    label: str = ''

    def __repr__(self) -> str:
        return f"('{self.term.name}', '{self.label}')"


@dataclass
class Function:
    name: str
    label: str
    input_types: list[FunctionArgument]
    output_type: FunctionArgument
    attributes: FunctionAttributes = field(default_factory=FunctionAttributes)

    def __repr__(self) -> str:
        return (
            f'Function(name={self.name}, label={self.label}, input_types={self.input_types},'
            f' output_type={self.output_type}, '
            f'attributes={self.attributes})'
        )

    def _format_input_types(self) -> str:
        return (
            '['
            + ', '.join(
                f'{{name: {t.term.name}, label: {t.label}}}' for t in self.input_types
            )
            + ']'
        )

    def _format_output_type(self) -> str:
        return (
            f'{{name: {self.output_type.term.name}, label: {self.output_type.label}}}'
        )


@dataclass
class RelationshipAttributes:
    color: Optional[str] = None
    direction: Optional[RelationshipDirection] = None
    title: Optional[str] = None
    rightChar: Optional[str] = None
    leftChar: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f'RelationshipAttributes(color={self.color}, '
            f'direction={self.direction.value if self.direction else self.direction}, '
            f'title={self.title}, '
            f'rightChar={self.rightChar}, '
            f'leftChar={self.leftChar})'
        )


@dataclass
class Relationship:
    parent: Term
    relationship: RelationshipType
    children: list[Term]
    name: Optional[str] = None
    attributes: RelationshipAttributes = field(default_factory=RelationshipAttributes)

    def __repr__(self) -> str:
        return (
            f'Relationship(parent={self.parent.name}, '
            f'relationship={self.relationship.value}, '
            f'children=[{",".join(child.name for child in self.children)}], attributes={self.attributes})'
        )


@dataclass
class Figure:
    name: str
    types: list[Term] = field(default_factory=list)
    functions: list[Function] = field(default_factory=list)
    hierarchy: list[Relationship] = field(default_factory=list)

    def __repr__(self) -> str:
        return f'Figure(name={self.name}, tyoes={self.types}, functions={self.functions}, hierarchy={self.hierarchy})'


@dataclass
class Ontology:
    meta: Meta = field(default_factory=Meta)
    types: list[Term] = field(default_factory=list)
    functions: list[Function] = field(default_factory=list)
    hierarchy: list[Relationship] = field(default_factory=list)
    figures: list[Figure] = field(default_factory=list)

    @staticmethod
    def from_figure(parent_ontology: 'Ontology', figure: Figure) -> 'Ontology':
        return Ontology(
            meta=parent_ontology.meta.with_new_name(figure.name),
            types=figure.types,
            functions=figure.functions,
            hierarchy=figure.hierarchy,
        )

    def add_type(self, type_def: Term) -> None:
        self.types.append(type_def)

    def add_function(self, func_def: Function) -> None:
        self.functions.append(func_def)

    def add_relationship(self, relationship: Relationship) -> None:
        self.hierarchy.append(relationship)

    def add_figure(self, figure: Figure) -> None:
        self.figures.append(figure)

    def set_meta(self, meta: Meta) -> None:
        self.meta = meta

    def find_term_by_name(self, name: str) -> Optional[Term]:
        return next((term for term in self.types if term.name == name), None)

    def find_function_by_name(self, name: str) -> Optional[Function]:
        return next(
            (function for function in self.functions if function.name == name), None
        )

    def find_relationship_by_name(self, name: str) -> Optional[Relationship]:
        return next(
            (
                relationship
                for relationship in self.hierarchy
                if relationship.name == name
            ),
            None,
        )

    def find_definition_by_name(
        self, name: str
    ) -> Optional[Term | Function | Relationship]:
        definitions: list[Term | Function | Relationship] = (
            self.types + self.functions + self.hierarchy
        )
        return next(
            (definition for definition in definitions if definition.name == name),
            None,
        )

    @property
    def without_functions(self) -> 'Ontology':
        types_used_in_hierarchy = [
            term
            for relationship in self.hierarchy
            for term in [relationship.parent] + relationship.children
        ]
        return Ontology(
            meta=self.meta.with_new_name(f'{self.meta.title} | Иерархия'),
            types=types_used_in_hierarchy,
            functions=[],
            hierarchy=self.hierarchy,
            figures=self.figures,
        )

    @property
    def only_functions(self) -> 'Ontology':
        types_used_in_functions = [
            arg.term
            for func in self.functions
            for arg in func.input_types + [func.output_type]
        ]
        return Ontology(
            meta=self.meta.with_new_name(f'{self.meta.title} | Алгоритмы'),
            types=types_used_in_functions,
            functions=self.functions,
            hierarchy=[],
            figures=self.figures,
        )

    def __repr__(self) -> str:
        return (
            f'Ontology(meta={self.meta}, '
            f'types={self.types}, '
            f'functions={self.functions}, '
            f'hierarchy={self.hierarchy}, '
            f'figures={self.figures})'
        )

    def count_edges(self) -> int:
        count: int = len(self.hierarchy)
        for func in self.functions:
            types: list[str] = [attribute.term.name for attribute in func.input_types]
            count += len(set(types)) + 1
        return count
